from .validators import Validator
from .errors import ValidationError


class ValidatorsNullCombination(object):
    """Used as a default validators combination when a field has no
    validators.
    """

    def __call__(self, *args, **kwargs):
        """Null combination does nothing"""

    def __and__(self, other):
        return ValidatorsAndCombination([other])

    def __or__(self, other):
        return ValidatorsOrCombination([other])


class ValidatorsAndCombination(object):
    """Represents an & combination of validators.
    It raise error if at least one validator is invalid.
    """
    def __init__(self, validators):
        """
        Args:
            validators list<Validator>: a list of validator
        """
        self._validators = validators

    def __call__(self, *args, **kwargs):
        for validator in self._validators:
            validator(*args, **kwargs)

    def __and__(self, other):
        if isinstance(other, ValidatorsAndCombination):
            self._validators.extend(other._validators)
            return self
        elif isinstance(other, Validator):
            self._validators.append(other)
            return self

        return ValidatorsAndCombination([self, other])

    def __or__(self, other):
        return ValidatorsOrCombination([self, other])


class ValidatorsOrCombination(object):
    """Represents an | combination of validators.
    It raise error if all validators are invalid.
    """
    def __init__(self, validators):
        """
        Args:
            validators list<Validator>: a list of validator
        """
        self._validators = validators

    def __call__(self, *args, **kwargs):
        error = None
        for validator in self._validators:
            try:
                validator(*args, **kwargs)
                return
            except ValidationError as err:
                error = err

        raise error

    def __and__(self, other):
        return ValidatorsAndCombination([self, other])

    def __or__(self, other):
        if isinstance(other, ValidatorsOrCombination):
            self._validators.extend(other._validators)
            return self
        elif isinstance(other, Validator):
            self._validators.append(other)
            return self

        return ValidatorsOrCombination([self, other])



from . import Field
from .compatibility import merge_validator
from ..validators import Type


class Bool(Field):
    """Field used to handle boolean values"""

    def __init__(self, **kwargs):
        merge_validator(kwargs, Type(bool))
        super(Bool, self).__init__(**kwargs)



import pytest

from lupin.validators import IsNone
from lupin.errors import NotNone


@pytest.fixture
def validator():
    return IsNone()


class TestCall(object):
    def test_raise_error_if_not_none(self, validator):
        with pytest.raises(NotNone):
            validator("46", [])

    def test_does_nothing_if_none(self, validator):
        validator(None, [])



from . import Validator
from ..errors import InvalidMatch


class Match(Validator):
    """Validate that a string matches a pattern"""

    def __init__(self, regex):
        """
        Args:
            regex (regex): a regexp object
        """
        self._regex = regex

    def __call__(self, value, path):
        """Validate that value matches the regex

        Args:
            value (str): string to validate
            path (list): error path
        """
        try:
            if not self._regex.match(value):
                raise InvalidMatch(value, self._regex, path)
        except TypeError:
            raise InvalidMatch(value, self._regex, path)



from . import Field
from .compatibility import merge_validator
from ..validators import Equal


class Constant(Field):
    """Field used to display a fixed value"""

    def __init__(self, value, **kwargs):
        """
        Args:
            value (object): fixed value
        """
        merge_validator(kwargs, Equal(value))
        super(Constant, self).__init__(**kwargs)
        self._value = value

    def load(self, value, mapper):
        """Returns fixed value

        Args:
            value (object): ignored
            mapper (Mapper): mapper used to load data

        Returns:
            object
        """
        return self._value

    def dump(self, value, mapper):
        """Returns fixed value

        Args:
            value (object): ignored
            mapper (Mapper): mapper used to dump data

        Returns:
            object
        """
        return self._value

    def extract_attr(self, obj, mapper, key=None):
        """Returns fixed value

        Args:
            obj (object): object to get value from
            mapper (Mapper): mapper used to dump data
            key (str): attribute name

        Returns:
            object
        """
        return self._value



from lupin import ValidatorsNullCombination, ValidatorsOrCombination, ValidatorsAndCombination
from lupin.validators import IsNone
import pytest


@pytest.fixture
def combination():
    return ValidatorsNullCombination()


class TestAnd(object):
    def test_returns_an_and_combination(self, combination):
        result = combination & IsNone()
        assert isinstance(result, ValidatorsAndCombination)


class TestOr(object):
    def test_returns_an_or_combination(self, combination):
        result = combination | IsNone()
        assert isinstance(result, ValidatorsOrCombination)



import pytest

from lupin import PolymorphicList, Mapper
from lupin.errors import InvalidType, InvalidDocument, MissingPolymorphicKey,\
    InvalidPolymorphicType

from tests.fixtures import Jewel, Painting


@pytest.fixture
def mapper():
    return Mapper()

@pytest.fixture
def field(mapper, jewel_schema, painting_schema):
    mapper.register(Jewel, jewel_schema)
    mapper.register(Painting, painting_schema)
    return PolymorphicList(on="type",
                           schemas={
                               "jewel": jewel_schema,
                               "painting": painting_schema
                           })


class TestLoad(object):
    def test_loads_different_object_types(self, field, stolen_items_data, mapper):
        result = field.load(stolen_items_data, mapper)
        assert len(result) == 2
        mona_lisa, diamond = result
        assert isinstance(mona_lisa, Painting)
        assert isinstance(diamond, Jewel)

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.load(None, mapper) is None


class TestDump(object):
    def test_dumps_different_object_types(self, field, stolen_items, stolen_items_data, mapper):
        result = field.dump(stolen_items, mapper)
        assert result == stolen_items_data

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.dump(None, mapper) is None


class TestValidate(object):
    def test_do_not_raise_error_if_none_and_allow_none(self, mapper, jewel_schema, painting_schema):
        mapper.register(Jewel, jewel_schema)
        mapper.register(Painting, painting_schema)
        field = PolymorphicList(on="type",
                                schemas={
                                    "jewel": jewel_schema,
                                    "painting": painting_schema
                                },
                                allow_none=True)
        field.validate(None, [], mapper)

    def test_raise_exception_if_not_list(self, field, mapper):
        with pytest.raises(InvalidType):
            field.validate("", [], mapper)

    def test_raise_exception_if_an_item_is_invalid(self, field, stolen_items_data, mapper):
        stolen_items_data[1]["carat"] = "20"
        with pytest.raises(InvalidDocument) as exc:
            field.validate(stolen_items_data, [], mapper)

        errors = exc.value
        error = errors[0]
        assert error.path == ["1", "carat"]

    def test_raises_exception_if_no_polymorphic_key(self, field, stolen_items_data, mapper):
        del stolen_items_data[0]["type"]
        with pytest.raises(MissingPolymorphicKey):
            field.validate(stolen_items_data, [], mapper)

    def test_raises_exception_if_invalid_polymorphic_type(self, field, stolen_items_data, mapper):
        stolen_items_data[0]["type"] = "car"
        with pytest.raises(InvalidPolymorphicType):
            field.validate(stolen_items_data, [], mapper)

    def test_does_nothing_if_all_items_are_valid(self, field, stolen_items_data, mapper):
        field.validate(stolen_items_data, [], mapper)



from copy import copy
from . import bind
from .errors import ValidationError, InvalidDocument, MissingKey


# compatibility : used to generate schema names if not provided (2018-07-04)
_globals = {
    "schemas_count": 0
}


def _generate_name():
    """Generate schema name if not provided"""
    _globals["schemas_count"] += 1
    return "schema%i" % _globals["schemas_count"]


class Schema(object):
    def __init__(self, fields, name=None):
        """
        Args:
            fields (dict): dictionary of fields
            name (str): schema name
        """
        self._fields = fields
        if not name:
            name = _generate_name()
        self.name = name

    def add_field(self, name, field):
        """Add new field to schema.

        Args:
            name (str): field name
            field (Field): a field
        """
        self._fields[name] = field

    def copy(self, new_name=None):
        """Returns a new schema based on current schema

        Args:
            new_names (str): name of new schema

        Returns:
            Schema
        """
        return type(self)(copy(self._fields), new_name)

    def load(self, cls, data, mapper, allow_partial=False, factory=bind):
        """Loads an instance of cls from dictionary

        Args:
            cls (class): class to instantiate
            data (dict): dictionary of data
            mapper (Mapper): mapper used to load data
            allow_partial (bool): allow partial schema, won't raise error if missing keys
            factory (callable): factory method used to instantiate objects

        Returns:
            object
        """
        attrs = self.load_attrs(data, mapper, allow_partial)
        return factory(cls, attrs)

    def load_attrs(self, data, mapper, allow_partial=False):
        """Loads attributes dictionary from `data`

        Args:
            data (dict): dictionary of data
            mapper (Mapper): mapper used to load data
            allow_partial (bool): allow partial schema, won't raise error if missing keys

        Returns:
            dict
        """
        attrs = {}
        for key, field in self._fields.items():
            if field.is_read_only:
                continue
            if key in data:
                value = field.load(data[key], mapper)
            elif allow_partial:
                continue
            else:
                value = field.default

            attr_name = field.binding or key
            attrs[attr_name] = value

        return attrs

    def dump(self, obj, mapper):
        """Dumps object into a dictionnary

        Args:
            obj (object): object to dump
            mapper (Mapper): mapper used to dump data

        Returns:
            dict
        """
        return {key: field.extract_attr(obj, mapper, key)
                for key, field
                in self._fields.items()}

    def validate(self, data, mapper, allow_partial=False, path=None):
        """Validate data with all field validators.
        If path is provided it will be used as the base path for errors.

        Args:
            data (dict): data to validate
            mapper (Mapper): mapper used to validate data
            allow_partial (bool): allow partial schema, won't raise error if missing keys
            path (list): base path for errors
        """
        path = path or []
        errors = []
        for key, field in self._fields.items():
            field_path = path + [key]
            if key in data:
                raw_value = data.get(key)
                try:
                    field.validate(raw_value, field_path, mapper)
                except ValidationError as error:
                    errors.append(error)
            elif not allow_partial and not field.is_optional:
                errors.append(MissingKey(key, field_path))

        if errors:
            raise InvalidDocument(errors)



from datetime import datetime
from lupin import fields, Mapper
import pytest


@pytest.fixture
def mapper():
    return Mapper()


@pytest.fixture
def field():
    return fields.DateTime()


@pytest.fixture
def time():
    return datetime(2017, 10, 7, 17, 35, 0)


class TestDump(object):
    def test_returns_datetime_as_string(self, field, time, mapper):
        assert field.dump(time, mapper) == "2017-10-07T17:35:00"

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.dump(None, mapper) is None

class TestLoad(object):
    def test_returns_a_datetime(self, field, time, mapper):
        assert field.load("2017-10-07T17:35:00", mapper) == time

    def test_returns_none_if_value_is_null(self, field, mapper):
        assert field.load(None, mapper) is None



import pytest

from lupin.validators import DateTimeFormat
from lupin.errors import InvalidDateTimeFormat


@pytest.fixture
def validator():
    return DateTimeFormat("%Y-%m-%d")


class TestCall(object):
    def test_raise_error_if_invalid_format(self, validator):
        with pytest.raises(InvalidDateTimeFormat):
            validator("01-01-2017", [])

    def test_does_nothing_if_valid_format(self, validator):
        validator("2017-01-01", [])



import warnings
from lupin import ValidatorsAndCombination


def merge_validator(kwargs, validator):
    """Add validator to field constructor parameters
    This method will be removed from next release.

    Args:
        kwargs (dict): constructor parameters
        validator (ValidatorsCombination): a validator or a combination
    """
    validators = kwargs.get("validators")
    if not validators:
        validators = validator
    elif isinstance(validators, list):
        warnings.warn("List of validators is deprecated, please use combinations (&|)", DeprecationWarning)
        validators.append(validator)
        validators = ValidatorsAndCombination(validators)
    else:
        validators = validators & validator

    kwargs["validators"] = validators



# coding: utf-8
import pytest

from lupin import Schema, String, Constant, Int, Mapping

from ..fixtures import Thief, Painting, Jewel


@pytest.fixture
def thief_schema():
    return Schema({
        "firstName": String(binding="first_name"),
        "lastName": String(binding="last_name")
    }, "thief")


@pytest.fixture
def painting_schema():
    return Schema({
        "author": String(),
        "name": String(),
        "type": Constant("painting")
    }, "painting")


@pytest.fixture
def jewel_schema():
    return Schema({
        "carat": Int(),
        "name": String(),
        "type": Constant("jewel")
    }, "jewel")


@pytest.fixture
def thief_data():
    return {
        "lastName": "Lupin",
        "firstName": "Arsène"
    }


@pytest.fixture
def thief():
    return Thief(last_name="Lupin",
                 first_name="Arsène")


@pytest.fixture
def mona_lisa():
    return Painting(name="Mona Lisa",
                    author="Leonardo da Vinci")


@pytest.fixture
def painting_mapping(painting_schema):
    return Mapping(Painting, painting_schema)


@pytest.fixture
def jewel_mapping(jewel_schema):
    return Mapping(Jewel, jewel_schema)


@pytest.fixture
def mona_lisa_data():
    return {
        "name": "Mona Lisa",
        "author": "Leonardo da Vinci",
        "type": "painting"
    }

@pytest.fixture
def diamond():
    return Jewel(name="diamond", carat=20)


@pytest.fixture
def diamond_data():
    return {
        "type": "jewel",
        "name": "diamond",
        "carat": 20
    }


@pytest.fixture
def stolen_items(mona_lisa, diamond):
    return [mona_lisa, diamond]


@pytest.fixture
def stolen_items_data(mona_lisa_data, diamond_data):
    return [mona_lisa_data, diamond_data]
